import ast
import typing

from sqlalchemy import and_, or_, select


class PyToSQLException(Exception):
    pass


class PyToSQLParsingError(PyToSQLException):
    pass


class _QueryVisitor(ast.NodeVisitor):
    def __init__(self, table):
        self.table = table
        self.conditions = []

    def _get_sides_of_compare(self, node: ast.Compare):
        return node.left, node.comparators[0]

    def _get_field(self, node: ast.Compare) -> str:
        for possible in self._get_sides_of_compare(node):
            if isinstance(possible, ast.Name):
                return possible.id
        raise PyToSQLParsingError(f"Node {node} does not have a name")

    def _get_value(self, node: ast.Compare) -> str:
        for possible in self._get_sides_of_compare(node):
            if isinstance(possible, ast.Constant):
                return possible.value
        raise PyToSQLParsingError(f"Node {node} does not have a value")

    def generic_visit(self, node):
        if not isinstance(node, (ast.Expression, ast.BoolOp, ast.Or, ast.And, ast.Constant)):
            raise PyToSQLParsingError(f"Unsupported node {node}")
        super().generic_visit(node)

    def visit_BoolOp(self, node: ast.BoolOp):
        if isinstance(node.op, ast.Or):
            op = or_
        elif isinstance(node.op, ast.And):
            op = and_
        else:
            raise PyToSQLParsingError(f"Unsupported bool operation {node.op}")
        self.generic_visit(node)
        condition = op(*self.conditions)
        self.conditions = [condition]

    def visit_Compare(self, node: ast.Compare):
        field = self._get_field(node)
        column = getattr(self.table, field)
        value = self._get_value(node)
        if isinstance(node.ops[0], ast.Eq):
            condition = column == value
        elif isinstance(node.ops[0], ast.NotEq):
            condition = column != value
        elif isinstance(node.ops[0], ast.In):
            condition = column.any(name=value)
        elif isinstance(node.ops[0], ast.NotIn):
            condition = ~column.any(name=value)
        else:
            raise PyToSQLParsingError(f"Unsupported operation {node.ops[0]}")
        self.conditions.append(condition)


def python_to_sqlalchemy_conditions(table, query):
    try:
        tree = ast.parse(query, mode="eval")
    except SyntaxError as e:
        raise PyToSQLParsingError(f"Invalid syntax in query `{query}`: {e.msg}") from e
    visitor = _QueryVisitor(table)
    visitor.visit(tree)
    return visitor.conditions


def python_to_sqlalchemy(table, query):
    return select(table).where(*python_to_sqlalchemy_conditions(table, query))



# pylint: disable=redefined-outer-name,unused-argument

import pytest
from sqlalchemy import Column, ForeignKey, Integer, String, Table, create_engine
from sqlalchemy.orm import Session, declarative_base, relationship  # pylint: disable=no-name-in-module

import pytosql

Base = declarative_base()

resources_labels = Table(
    "resources_labels",
    Base.metadata,
    Column("resource_id", ForeignKey("resources.id")),
    Column("label_id", ForeignKey("labels.id")),
)


class Resource(Base):
    __tablename__ = "resources"

    id = Column(Integer, primary_key=True)
    name = Column(String)
    labels = relationship("Label", secondary=resources_labels)

    def __repr__(self):
        return f"Resource(id={self.id!r}, name={self.name!r})"


class Label(Base):
    __tablename__ = "labels"

    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False)

    def __repr__(self):
        return f"Label(id={self.id!r}, name={self.name!r})"


@pytest.fixture
def session():
    engine = create_engine("sqlite://", echo=True, future=True)
    Base.metadata.create_all(engine)
    with Session(engine) as session:  # pylint: disable=not-context-manager
        yield session


@pytest.fixture
def labels(session):
    labels = [Label(name="L1"), Label(name="L2"), Label(name="C")]
    session.add_all(labels)
    session.commit()
    return labels


@pytest.fixture
def resources(session, labels):
    resources = [
        Resource(name="R1", labels=[labels[0], labels[2]]),
        Resource(name="R2", labels=[labels[1], labels[2]]),
    ]
    session.add_all(resources)
    session.commit()
    return resources


@pytest.mark.parametrize(
    "query,expected",
    [
        ("'L1' in labels", ["R1"]),
        ("'L2' in labels", ["R2"]),
        ("'C' in labels", ["R1", "R2"]),
        ("'L1' not in labels", ["R2"]),
        ("'L2' not in labels", ["R1"]),
        ("'C' not in labels", []),
        ("name == 'R1'", ["R1"]),
        ("name == 'R2'", ["R2"]),
        ("name != 'R1'", ["R2"]),
        ("name != 'R2'", ["R1"]),
        ("name == 'R1' and 'L1' in labels", ["R1"]),
        ("name == 'R1' and 'L2' in labels", []),
        ("name == 'R1' or 'L2' in labels", ["R1", "R2"]),
        ("(name == 'R1' and 'L1' in labels) or name == 'R2'", ["R1", "R2"]),
        ("(name == 'R1' and 'L2' in labels) or name == 'R2'", ["R2"]),
        (
            "(name == 'R1' and ('L2' in labels or 'C' in labels)) and 'L1' in labels",
            ["R1"],
        ),
    ],
)
def test_python_to_sqlalchemy(session, resources, query, expected):
    assert sorted(
        [resource.name for resource in session.scalars(pytosql.python_to_sqlalchemy(Resource, query))]
    ) == sorted(expected)


def test_syntax_error(session):
    expected = "Invalid syntax in query `name == 'hi`: "
    with pytest.raises(pytosql.PyToSQLParsingError, match=expected):
        pytosql.python_to_sqlalchemy(Resource, "name == 'hi")
